package org.gstreamer.encoding;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import org.gstreamer.Caps;
import org.gstreamer.GObject;
import org.gstreamer.discoverer.GstDiscovererInfo;
import org.gstreamer.lowlevel.GlibAPI.GList;
import org.gstreamer.lowlevel.GstEncodingProfileAPI;

/**
 *
 * encoding-profile
 *
 * encoding-profile — Encoding profile library
 *
 *
 * Properties GstCaps * restriction-caps Read / Write
 *
 * Description
 *
 * Functions to create and handle encoding profiles.
 *
 * Encoding profiles describe the media types and settings one wishes to use for
 * an encoding process. The top-level profiles are commonly
 * GstEncodingContainerProfile(s) (which contains a user-readable name and
 * description along with which container format to use). These, in turn,
 * reference one or more GstEncodingProfile(s) which indicate which encoding
 * format should be used on each individual streams.
 *
 * GstEncodingProfile(s) can be provided to the 'encodebin' element, which will
 * take care of selecting and setting up the required elements to produce an
 * output stream conforming to the specifications of the profile.
 *
 * Unlike other systems, the encoding profiles do not specify which GstElement
 * to use for the various encoding and muxing steps, but instead relies on
 * specifying the format one wishes to use.
 *
 * Encoding profiles can be created at runtime by the application or loaded from
 * (and saved to) file using the GstEncodingTarget API. The encoding profile
 * serialization format
 *
 * This is the serialization format of a GstEncodingProfile.
 *
 * Internally the transcoding application uses GstEncodeBin.
 * gst-validate-transcoding-&GST_API_VERSION; uses its own serialization format
 * to describe the GstEncodeBin.profile property of the encodebin.
 *
 * The simplest serialized profile looks like:
 *
 * 1
 *
 *
 *
 * muxer_source_caps:videoencoder_source_caps:audioencoder_source_caps
 *
 * For example to encode a stream into a WebM container, with an OGG audio
 * stream and a VP8 video stream, the serialized GstEncodingProfile will look
 * like:
 *
 * 1
 *
 *
 *
 * video/webm:video/x-vp8:audio/x-vorbis
 *
 * You can also set the preset name of the encoding profile using the
 * caps+preset_name syntax as in:
 *
 * 1
 *
 *
 *
 * video/webm:video/x-vp8+youtube-preset:audio/x-vorbis
 *
 * Moreover, you can set the presence property of an encoding profile using the
 * |presence syntax as in:
 *
 * 1
 *
 *
 *
 * video/webm:video/x-vp8|1:audio/x-vorbis
 *
 * This field allows you to specify how many times maximum a GstEncodingProfile
 * can be used inside an encodebin.
 *
 * You can also use the restriction_caps->encoded_format_caps syntax to specify
 * the restriction caps to be set on a GstEncodingProfile. It corresponds to the
 * restriction GstCaps to apply before the encoder that will be used in the
 * profile. The fields present in restriction caps are properties of the raw
 * stream (that is, before encoding), such as height and width for video and
 * depth and sampling rate for audio. This property does not make sense for
 * muxers.
 *
 * To force a video stream to be encoded with a Full HD resolution (using WebM
 * as the container format, VP8 as the video codec and Vorbis as the audio
 * codec), you should use:
 *
 * 1
 *
 *
 *
 * video/webm:video/x-raw,width=1920,height=1080->video/x-vp8:audio/x-vorbis
 *
 * Some serialized encoding formats examples:
 *
 * MP3 audio and H264 in MP4:
 *
 * 1
 *
 *
 *
 * video/quicktime,variant=iso:video/x-h264:audio/mpeg,mpegversion=1,layer=3
 *
 * Vorbis and theora in OGG:
 *
 * 1
 *
 *
 *
 * application/ogg:video/x-theora:audio/x-vorbis
 *
 * AC3 and H264 in MPEG-TS:
 *
 * 1
 *
 *
 *
 * video/mpegts:video/x-h264:audio/x-ac3
 */
/**
 *  Example: Creating a profile
*
* #include <gst/pbutils/encoding-profile.h> ... GstEncodingProfile *
* create_ogg_theora_profile(void) { GstEncodingContainerProfile *prof; GstCaps
* *caps;
*
* caps = gst_caps_from_string("application/ogg"); prof =
* gst_encoding_container_profile_new("Ogg audio/video",
* "Standard OGG/THEORA/VORBIS", caps, NULL); gst_caps_unref (caps);
*
* caps = gst_caps_from_string("video/x-theora");
* gst_encoding_container_profile_add_profile(prof, (GstEncodingProfile*)
* gst_encoding_video_profile_new(caps, NULL, NULL, 0)); gst_caps_unref (caps);
*
* caps = gst_caps_from_string("audio/x-vorbis");
* gst_encoding_container_profile_add_profile(prof, (GstEncodingProfile*)
* gst_encoding_audio_profile_new(caps, NULL, NULL, 0)); gst_caps_unref (caps);
*
* return (GstEncodingProfile*) prof; }
*/
/**
 * Example: Using an encoder preset with a profile
 *
 * #include <gst/pbutils/encoding-profile.h> ... GstEncodingProfile *
 * create_ogg_theora_profile(void) { GstEncodingVideoProfile *v;
 * GstEncodingAudioProfile *a; GstEncodingContainerProfile *prof; GstCaps *caps;
 * GstPreset *preset;
 *
 * caps = gst_caps_from_string ("application/ogg"); prof =
 * gst_encoding_container_profile_new ("Ogg audio/video",
 * "Standard OGG/THEORA/VORBIS", caps, NULL); gst_caps_unref (caps);
 *
 * preset = GST_PRESET (gst_element_factory_make ("theoraenc", "theorapreset")); // The preset will be saved on the filesystem,
 *              g_object_set (preset, "bitrate", 1000, NULL);  // so try to use a descriptive name gst_preset_save_preset
 *                                                              // (preset, "theora_bitrate_preset");
 * gst_object_unref (preset);
 *
 * caps = gst_caps_from_string ("video/x-theora"); v =
 * gst_encoding_video_profile_new (caps, "theorapreset", NULL, 0);
 * gst_encoding_container_profile_add_profile (prof, (GstEncodingProfile*) v);
 * gst_caps_unref (caps);
 *
 * caps = gst_caps_from_string ("audio/x-vorbis"); a =
 * gst_encoding_audio_profile_new (caps, NULL, NULL, 0);
 * gst_encoding_container_profile_add_profile (prof, (GstEncodingProfile*) a);
 * gst_caps_unref (caps);
 *
 * return (GstEncodingProfile*) prof; }
 */
/***
 * Example: Listing categories, targets and profiles
 *
 *
 * #include <gst/pbutils/encoding-profile.h> ... GstEncodingProfile *prof;
 * GList*categories, *tmpc; GList *targets, *tmpt; ... categories =
 * gst_encoding_list_available_categories ();
 *
 * ... Show available categories to user ...
 *
 * for (tmpc = categories; tmpc; tmpc = tmpc->next) { gchar *category = (gchar
 * *) tmpc->data;
 *
 * ... and we can list all targets within that category ...
 *
 * targets = gst_encoding_list_all_targets (category);
 *
 * ... and show a list to our users ...
 *
 * g_list_foreach (targets, (GFunc) gst_encoding_target_unref, NULL);
 * g_list_free (targets); }
 *
 * g_list_foreach (categories, (GFunc) g_free, NULL); g_list_free (categories);
 */

public class EncodingProfile extends GObject {
    public static final String GTYPE_NAME = "GstEncodingProfile";

    protected static final GstEncodingProfileAPI gst = GstEncodingProfileAPI.GSTENCODINGPROFILE_API;
    public EncodingProfile(Initializer init) {
        super(init);
    }

    /**
     * Find the GstEncodingProfile with the specified name and category.
     * Parameters
     *
     * targetname
     *
     *
     * The name of the target. [transfer none]
     *
     * profilename
     *
     * (allow-none): The name of the profile, if NULL provided, it will default
     * to the encoding profile called default. [transfer none]
     *
     * category
     *
     *
     * The target category. Can be NULL. [transfer none][allow-none] Returns
     *
     * The matching GstEncodingProfile or NULL.
     *
     * [transfer full]
     */
    public static EncodingProfile find(String targetname, String profilename, String category) {
        return gst.gst_encoding_profile_find(targetname, profilename, category);
    }

    /**
     * Creates a GstEncodingProfile matching the formats from the given
     * GstDiscovererInfo. Streams other than audio or video (eg, subtitles), are
     * currently ignored. Parameters
     *
     * info
     *
     *
     * The GstDiscovererInfo to read from. [transfer none] Returns
     *
     * The new GstEncodingProfile or NULL.
     *
     * [transfer full]
     */
    public static EncodingProfile fromDiscoverer(GstDiscovererInfo info) {
        return gst.gst_encoding_profile_from_discoverer(info);
    }

    /**
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * the name of the profile, can be NULL.
     * gst_encoding_profile_get_description ()
     *
     * const gchar * gst_encoding_profile_get_description (GstEncodingProfile
     * *profile);
     *
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * the description of the profile, can be NULL.
     */
    public String getName() {
        return gst.gst_encoding_profile_get_name(this);
    }


    /**
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * the GstCaps corresponding to the media format used in the profile. Unref
     * after usage. [transfer full]
     */
    public Caps getFormat() {
        Caps caps = gst.gst_encoding_profile_get_format(this);

        return caps;
    }

    /**
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * the name of the GstPreset to be used in the profile. This is the name
     * that has been set when saving the preset.
     * gst_encoding_profile_get_preset_name ()
     */

    public String getPreset() {
        return gst.gst_encoding_profile_get_preset(this);
    }

    /**
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * the name of the GstPreset factory to be used in the profile.
     * gst_encoding_profile_get_presence ()
     */
    public String getPresetName() {
        return gst.gst_encoding_profile_get_preset_name(this);
    }

    /**
     *
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * The number of times the profile is used in its parent container profile.
     * If 0, it is not a mandatory stream. gst_encoding_profile_get_restriction
     * ()
     */

    public int getPresence() {
        return gst.gst_encoding_profile_get_presence(this);
    }

    /**
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * The restriction GstCaps to apply before the encoder that will be used in
     * the profile. The fields present in restriction caps are properties of the
     * raw stream (that is before encoding), such as height and width for video
     * and depth and sampling rate for audio. Does not apply to
     * GstEncodingContainerProfile (since there is no corresponding raw stream).
     * Can be NULL. Unref after usage.
     *
     * [transfer full]
     */
    public Caps getRestriction() {
        return gst.gst_encoding_profile_get_restriction(this);
    }

    /**
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * a suitable file extension for profile , or NULL.
     * gst_encoding_profile_set_name ()
     */
    public String getFileExtension() {
        return gst.gst_encoding_profile_get_file_extension(this);
    }

    /**
     * Set name as the given name for the profile . A copy of name will be made
     * internally. Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * name
     *
     *
     * the name to set on the profile
     */

    public void setName(String name) {
        gst.gst_encoding_profile_set_name(this, name);
    }

    /**
     * gst_encoding_profile_set_description ()
     *
     * void gst_encoding_profile_set_description (GstEncodingProfile *profile,
     * const gchar *description);
     *
     * Set description as the given description for the profile . A copy of
     * description will be made internally. Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * description
     *
     *
     * the description to set on the profile
     */
    public void setDescription(String name) {
        gst.gst_encoding_profile_set_description(this, name);
    }

    /**
     * Set whether the profile should be used or not. Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * enabled
     *
     *
     * FALSE to disable profile, TRUE to enable it
     */
    public void setEnabled(boolean enabled) {
        gst.gst_encoding_profile_set_enabled(this, enabled);
    }

    /**
     *
     * Sets the media format used in the profile. Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * format
     *
     *
     * the media format to use in the profile.
     */
    public void setFormat(Caps format) {
        gst.gst_encoding_profile_set_format(this, format);
    }

    /**
     *
     * Sets the name of the GstElement that implements the GstPreset interface
     * to use for the profile. This is the name that has been set when saving
     * the preset. Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * preset
     *
     *
     * the element preset to use
     */
    public void setPreset(EncodingProfile profile, String preset) {
        gst.gst_encoding_profile_set_preset(this, preset);
    }


    /**
     * gst_encoding_profile_set_preset_name ()
     *
     * void gst_encoding_profile_set_preset_name (GstEncodingProfile *profile,
     * const gchar *preset_name);
     *
     * Sets the name of the GstPreset's factory to be used in the profile.
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * preset_name
     *
     *
     * The name of the preset to use in this profile .
     */
    public void setPresetName(String presetName) {
        gst.gst_encoding_profile_set_preset_name(this, presetName);
    }
    /**
     *
     * Set the restriction GstCaps to apply before the encoder that will be used
     * in the profile. See gst_encoding_profile_get_restriction() for more about
     * restrictions. Does not apply to GstEncodingContainerProfile. Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * restriction
     *
     *
     * the restriction to apply. [transfer full]
     * gst_encoding_profile_set_presence ()
     */

    public void setRestriction(Caps restriction) {
        gst.gst_encoding_profile_set_restriction(this, restriction);
    }

    /**
     * Set the number of time the profile is used in its parent container
     * profile. If 0, it is not a mandatory stream Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     *
     * presence
     *
     *
     * the number of time the profile can be used
     */
    public void setPresence (int presence) {
        gst.gst_encoding_profile_set_presence(this, presence);
    }

    /**
     * gst_encoding_profile_is_equal ()
     *
     * Checks whether the two GstEncodingProfile are equal Parameters
     *
     * a
     *
     *
     * a GstEncodingProfile
     *
     *
     * b
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * TRUE if a and b are equal, else FALSE. gst_encoding_profile_is_enabled ()
     */
    public static boolean profileIsEqual(EncodingProfile a, EncodingProfile b) {
        // TODO: implement equals in base of this , and hashCode
        return gst.gst_encoding_profile_is_equal(a, b);
    }


    /**
     * boolean gst_encoding_profile_is_enabled (GstEncodingProfile *profile);
     */
    public boolean isEnabled() {
        return gst.gst_encoding_profile_is_enabled(this);
    }

    /**

     * Computes the full output caps that this profile will be able to consume.
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * The full caps the given profile can consume. Call gst_caps_unref() when
     * you are done with the caps.
     *
     * [transfer full]
     */
    public Caps getInputCaps () {
        return gst.gst_encoding_profile_get_input_caps(this);
    }

    /**
     * gst_encoding_profile_get_type_nick ()
     *
     * Parameters
     *
     * profile
     *
     *
     * a GstEncodingProfile
     *
     * Returns
     *
     * the human-readable name of the type of profile .
     */

    public String getTypeNick() {
        return gst.gst_encoding_profile_get_type_nick(this);
    }

    /**
     * List all available GstEncodingTarget for the specified category, or all
     * categories if categoryname is NULL. Parameters
     *
     * categoryname
     *
     *
     * The category, for ex: GST_ENCODING_CATEGORY_DEVICE. Can be NULL.
     * [allow-none] Returns
     *
     * The list of GstEncodingTarget.
     *
     * [transfer full][element-type GstEncodingTarget]
     */
    public static List<EncodingTarget> listAllTargets(String categoryname) {
        GList glist = gst.gst_encoding_list_all_targets(categoryname);
        return GList.convertToListOfType(glist, EncodingTarget.class);
    }

    /**
     * Lists all GstEncodingTarget categories present on disk. Returns
     *
     * A list of GstEncodingTarget categories.
     *
     * [transfer full][element-type gchar*]
     */
    public static List<String> listAvailableCategories() {
        GList glist = gst.gst_encoding_list_available_categories();
        List<String> list = new ArrayList<String>();
        GList next = glist;
        while (Objects.nonNull(next)) {
            if (Objects.nonNull(next.data)) {
                list.add(next.data.getString(0));
            }
            GList old = next;
            next = next.next();
            // ges.g_list_free();
        }

        return list;
    }

     /*
     *
     *
     * /* Types and Values
     * GstEncodingProfile
     *
     * typedef struct _GstEncodingProfile GstEncodingProfile;
     *
     * The opaque base class object for all encoding profiles. This contains
     * generic information like name, description, format and preset.
     * GstEncodingContainerProfile
     *
     * typedef struct _GstEncodingContainerProfile GstEncodingContainerProfile;
     *
     * Encoding profiles for containers. Keeps track of a list of
     * GstEncodingProfile GstEncodingAudioProfile
     *
     * typedef struct _GstEncodingAudioProfile GstEncodingAudioProfile;
     *
     * Variant of GstEncodingProfile for audio streams. GstEncodingVideoProfile
     *
     * typedef struct _GstEncodingVideoProfile GstEncodingVideoProfile;
     *
     * Variant of GstEncodingProfile for video streams, allows specifying the
     * pass . GST_ENCODING_CATEGORY_DEVICE
     *
     * #define GST_ENCODING_CATEGORY_DEVICE "device"
     *
     * GstEncodingTarget category for device-specific targets. The name of the
     * target will usually be the constructor and model of the device, and that
     * target will contain GstEncodingProfiles suitable for that device.
     * GST_ENCODING_CATEGORY_ONLINE_SERVICE
     *
     * #define GST_ENCODING_CATEGORY_ONLINE_SERVICE "online-service"
     *
     * GstEncodingTarget category for online-services. The name of the target
     * will usually be the name of the online service and that target will
     * contain GstEncodingProfiles suitable for that online service.
     * GST_ENCODING_CATEGORY_STORAGE_EDITING
     *
     * #define GST_ENCODING_CATEGORY_STORAGE_EDITING "storage-editing"
     *
     * GstEncodingTarget category for storage, archiving and editing targets.
     * Those targets can be lossless and/or provide very fast random access
     * content. The name of the target will usually be the container type or
     * editing target, and that target will contain GstEncodingProfiles suitable
     * for editing or storage. GST_ENCODING_CATEGORY_CAPTURE
     *
     * #define GST_ENCODING_CATEGORY_CAPTURE "capture"
     *
     * GstEncodingTarget category for recording and capture. Targets within this
     * category are optimized for low latency encoding. GstEncodingTarget
     *
     * typedef struct _GstEncodingTarget GstEncodingTarget;
     *
     * Collection of GstEncodingProfile for a specific target or use-case.
     *
     * When being stored/loaded, targets come from a specific category, like
     * GST_ENCODING_CATEGORY_DEVICE. Property Details The “restriction-caps”
     * property
     *
     * “restriction-caps” GstCaps *
     *
     * The restriction caps to use.
     *
     * Flags: Read / Write
     */

}
