package org.gstreamer.ges;

import java.util.List;

import org.gstreamer.ClockTime;
import org.gstreamer.GObject;
import org.gstreamer.ges.clip.GESClip;
import org.gstreamer.lowlevel.GESLayerAPI;
import org.gstreamer.lowlevel.GlibAPI;
import org.gstreamer.lowlevel.GlibAPI.GList;
import org.gstreamer.lowlevel.GstNative;

/**
 * GESLayer
 *
 * GESLayer—Non- overlapping sequence of GESClip
 *
 * Description
 *
 * Responsible for the ordering of the various contained
 * Clip(s). A timeline layer has a "priority" property, which is used to manage
 * the priorities of individual Clips. Two layers should not have the same
 * priority within a given timeline.
 */
/** Properties:
 *
 * boolean auto-transition Read / Write
 * guint priority Read
 */
/**
 * Write Signals
 *
 * void clip- added Run First void clip- removed Run First
 */
/**
 * Property Details The “auto-transition” property
 *
 *
 * “auto-transition” gboolean
 *
 * Sets whether transitions are added automagically when clips overlap.
 *
 * Flags: Read / Write
 *
 * Default value: FALSE The “priority” property
 *
 * “priority” guint
 *
 * The priority of the layer in the GESTimeline. 0 is the highest priority.
 * Conceptually, a GESTimeline is a stack of GESLayers, and the priority of the
 * layer represents its position in the stack. Two layers should not have the
 * same priority within a given GESTimeline.
 *
 * Note that the timeline needs to be
 *
 * commited (with ges_timeline_commit) for the change to be taken into account.
 *
 * Flags: Read / Write
 *
 * Default value: 0 Signal Details The “clip-added” signal
 *
 * void user_function (GESLayer *layer, GESClip *clip, gpointer user_data)
 *
 * Will be emitted after the clip was added to the layer. Parameters
 *
 * layer
 *
 *
 * the GESLayer
 *
 *
 * clip
 *
 *
 * the GESClip that was added.
 *
 *
 * user_data
 *
 *
 * user data set when the signal handler was connected.
 *
 *
 * Flags: Run First The “clip-removed” signal
 *
 * void user_function (GESLayer *layer, GESClip *clip, gpointer user_data)
 *
 * Will be emitted after the clip was removed from the layer. Parameters
 *
 * layer
 *
 *
 * the GESLayer
 *
 *
 * clip
 *
 *
 * the GESClip that was removed
 *
 *
 * user_data
 *
 *
 * user data set when the signal handler was connected.
 *
 *
 * Flags: Run First
 */
public class GESLayer extends GObject {
    public static final String GTYPE_NAME = "GESLayer";
    
    private static final GESLayerAPI ges = GstNative.load("ges", GESLayerAPI.class);

    public GESLayer(Initializer init) {
        super(init);
    }

    /**
     *
     * Adds the given clip to the layer. Sets the clip's parent, and thus takes
     * ownership of the clip.
     *
     * An clip can only be added to one layer.
     *
     * Calling this method will construct and properly set all the media related
     * elements on clip . If you need to know when
     *
     * those objects (actually GESTrackElement) are constructed, you should
     * connect to the container::child-added signal which is emited right after
     * those elements are ready to be used. Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     *
     * clip
     *
     *
     * the GESClip to add. [transfer full] Returns
     *
     * TRUE if the clip was properly added to the layer, or FALSE if the layer
     * refuses to add the clip.
     */
    public boolean addClip(GESClip clip) {
        return ges.ges_layer_add_clip(this, clip);
    }

    /**
     * Creates Clip from asset, adds it to layer and returns a reference to it.
     * Parameters
     *
     * layer - a GESLayer
     *
     * asset - The asset to add to
     *
     * start - The start value to set on the new GESClip, if start ==
     * GST_CLOCK_TIME_NONE, it will be set to the current duration of layer
     *
     * inpoint - The inpoint value to set on the new GESClip
     *
     * duration - The duration value to set on the new GESClip
     *
     * track_types - The GESTrackType to set on the the new GESClip
     *
     * Returns
     *
     * Created GESClip.
     *
     * [transfer none]
     */
    public GESClip addAsset(GESAsset asset, ClockTime start, ClockTime inpoint, ClockTime duration,
            int track_types) {
        return ges.ges_layer_add_asset(this, asset, start, inpoint, duration, track_types);
    }

    public GESClip addAsset(GESAsset asset, ClockTime start, ClockTime inpoint, ClockTime duration,
            GESTrackType track_type) {
        return ges.ges_layer_add_asset(this, asset, start, inpoint, duration, track_type.intValue());
    }

    /**
     * ges_layer_new ()
     *
     * GESLayer * ges_layer_new (void);
     *
     * Creates a new GESLayer. Returns A new GESLayer.
     */
    public GESLayer() {
        this(initializer(ges.ptr_ges_layer_new(), false, true));
    }

    /**
     *
     *
     * Removes the given clip from the layer and unparents it. Unparenting it
     * means the reference owned by layer on the clip will be removed. If you
     * wish to use the clip after this function, make sure you call
     * gst_object_ref() before removing it from the layer . Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     *
     * clip
     *
     *
     * the GESClip to remove
     *
     * Returns
     *
     * TRUE if the clip could be removed, FALSE if the layer does not want to
     * remove the clip. ges_layer_set_priority ()
     */
    public boolean removeClip(GESClip clip) {
        return ges.ges_layer_remove_clip(this, clip);
    }

    /**
     * Sets the layer to the given priority . See the documentation of the
     * priority property for more information. Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     *
     * priority
     *
     *
     * the priority to set
     */
    public void setPriority(int priority) {
        ges.ges_layer_set_priority(this, priority);

    }

    /**
     * Get the priority of layer within the timeline. Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     * Returns
     *
     * The priority of the layer within the timeline. ges_layer_get_clips ()
     */

    public int getPriority() {
        return ges.ges_layer_get_priority(this);
    }

    /**
     *
     *
     * Get the clips this layer contains. Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     * Returns
     *
     * a GList of clips. The user is responsible for unreffing the contained
     * objects and freeing the list.
     *
     * [transfer full][element-type GESClip]
     */
    public List<GESClip> getClips() {
        GList glist = ges.ges_layer_get_clips(this);
        List<GESClip> list = GList.convertToListOfType(glist, GESClip.class);
        GlibAPI.GLIB_API.g_list_free(glist);
        return list;
    }

    /**
     *
     * Get the GESTimeline in which GESLayer currently is. Parameters
     *
     * layer
     *
     *
     * The GESLayer to get the parent GESTimeline from
     *
     * Returns
     *
     * the GESTimeline in which GESLayer currently is or NULL if not in any
     * timeline yet.
     *
     * [transfer none]
     */
    GESTimeline getTimeline() {
        return ges.ges_layer_get_timeline(this);
    }

    /**
     *
     * Gets whether transitions are automatically added when objects overlap or
     * not. Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     * Returns
     *
     * TRUE if transitions are automatically added, else FALSE.
     * ges_layer_set_auto_transition ()
     */
    public boolean getAutoTransition() {
        return ges.ges_layer_get_auto_transition(this);
    }

    /**
     * Sets the layer to the given auto_transition . See the documentation of
     *
     * the property auto_transition for more information. Parameters
     *
     * layer
     *
     *
     * a GESLayer
     *
     *
     * auto_transition
     *
     *
     * whether the auto_transition is active
     */
    public void setAutoTransition(boolean auto_transition) {
        ges.ges_layer_set_auto_transition(this, auto_transition);
    }

    /**
     *
     *
     *
     *
     * Convenience method to check if layer is empty (doesn't contain any clip),
     * or not. Parameters
     *
     * layer
     *
     *
     * The GESLayer to check
     *
     * Returns
     *
     * TRUE if layer is empty, FALSE if it already contains at least one GESClip
     * ges_layer_get_duration ()
     */
    public boolean isEmpty() {
        return ges.ges_layer_is_empty(this);
    }

    /**
     * Lets you retrieve the duration of the layer, which means the end time of
     * the last clip inside it Parameters
     *
     * layer
     *
     *
     * The GESLayer to get the duration from
     *
     * Returns
     *
     * The duration of a layer GES_TIMELINE_GET_LAYERS()
     */
    public ClockTime getDuration(GESLayer layer) {
        return ges.ges_layer_get_duration(this);
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder('[');
        sb.append(super.toString()).append(",priority="+getPriority()).append(",isEmpty=").append(isEmpty());
        sb.append(",getAutoTransition=").append(getAutoTransition());
        sb.append(']');
        return sb.toString();
    }
}
