package org.gstreamer.ges;

import java.util.List;
import java.util.Objects;
import java.util.logging.Logger;

import org.gstreamer.Bin;
import org.gstreamer.ClockTime;
import org.gstreamer.GError;
import org.gstreamer.GstException;
import org.gstreamer.Pad;
import org.gstreamer.lowlevel.GESTimelineAPI;
import org.gstreamer.lowlevel.GlibAPI;
import org.gstreamer.lowlevel.GlibAPI.GList;
import org.gstreamer.lowlevel.GstAPI.GErrorStruct;

import com.sun.jna.Pointer;

/**
 * Description
 *
 * GESTimeline is the central object for any multimedia timeline.
 *
 * Contains a list of GESLayer which users should use to arrange the various
 * clips through time.
 *
 * The output type is determined by the GESTrack that are set on the
 * GESTimeline.
 *
 * To save/load a timeline, you can use the ges_timeline_load_from_uri() and
 * ges_timeline_save_to_uri() methods to use the default format. If you wish
 *
 * Note that any change you make in the timeline will not actually be taken into
 * account until you call the ges_timeline_commit method.
 */

/**
 * Property Details
 *
 * The “auto-transition” property
 *
 * “auto-transition” gboolean Sets whether transitions are added automagically
 * when clips overlap.
 *
 * Flags: Read / Write
 *
 * Default value: FALSE
 *
 * The “duration” property
 *
 * “duration” guint64 Current duration (in nanoseconds) of the GESTimeline
 *
 * Flags: Read
 *
 * Default value: 18446744073709551615
 *
 * The “snapping-distance” property
 *
 * “snapping-distance” guint64 Distance (in nanoseconds) from which a moving
 * object will snap with it neighboors. 0 means no snapping.
 *
 * Flags: Read / Write
 *
 * Default value: 0
 */

/**
 * Signal Details
 *
 * The “commited” signal
 *
 * void user_function (GESTimeline *timeline, gpointer user_data) This signal
 * will be emitted once the changes initiated by ges_timeline_commit have been
 * executed in the backend. Use ges_timeline_commit_sync if you don't need to do
 * anything in the meantime.
 *
 * Parameters
 *
 * timeline the GESTimeline user_data user data set when the signal handler was
 * connected. Flags: Run Last
 *
 * The “layer-added” signal
 *
 * void user_function (GESTimeline *timeline, GESLayer *layer, gpointer
 * user_data) Will be emitted after the layer was added to the timeline.
 *
 * Parameters
 *
 * timeline the GESTimeline layer the GESLayer that was added to the timeline
 * user_data user data set when the signal handler was connected. Flags: Run
 * First
 *
 * The “layer-removed” signal
 *
 * void user_function (GESTimeline *timeline, GESLayer *layer, gpointer
 * user_data) Will be emitted after the layer was removed from the timeline.
 *
 * Parameters
 *
 * timeline the GESTimeline layer the GESLayer that was removed from the
 * timeline user_data user data set when the signal handler was connected.
 * Flags: Run First
 *
 * The “select-tracks-for-object” signal
 *
 * GPtrArray* user_function (GESTimeline *timeline, GESClip *clip,
 * GESTrackElement *track-element, gpointer user_data) Parameters
 *
 * timeline the GESTimeline clip The GESClip on which track -element will land
 * track-element The GESTrackElement for which to choose the tracks it should
 * land into user_data user data set when the signal handler was connected.
 * Returns
 *
 * a GPtrArray of GESTrack-s where that object should be added.
 *
 * [transfer full][element-type GESTrack]
 *
 * Flags: Run Last
 *
 * The “snapping-ended” signal
 *
 * void user_function (GESTimeline *gestimeline, GESTrackElement *arg1,
 * GESTrackElement *arg2, guint64 arg3, gpointer user_data) Flags: Run Last
 *
 * The “snapping-started” signal
 *
 * void user_function (GESTimeline *gestimeline, GESTrackElement *arg1,
 * GESTrackElement *arg2, guint64 arg3, gpointer user_data) Flags: Run Last
 *
 * The “track-added” signal
 *
 * void user_function (GESTimeline *timeline, GESTrack *track, gpointer
 * user_data) Will be emitted after the track was added to the timeline.
 *
 * Parameters
 *
 * timeline the GESTimeline track the GESTrack that was added to the timeline
 * user_data user data set when the signal handler was connected. Flags: Run
 * First
 *
 * The “track-removed” signal
 *
 * void user_function (GESTimeline *timeline, GESTrack *track, gpointer
 * user_data) Will be emitted after the track was removed from the timeline.
 *
 * Parameters
 *
 * timeline the GESTimeline track the GESTrack that was removed from the
 * timeline user_data user data set when the signal handler was connected.
 * Flags: Run First
 *
 */
public class GESTimeline extends Bin {

    private static final GESTimelineAPI ges = GESTimelineAPI.GESTIMELINE_API;

    public GESTimeline(Initializer init) {
        super(init);
    }

    /**
     *
     * Creates a new empty GESTimeline.
     *
     * Returns
     *
     * The new timeline. ----------------------- GESTimeline *
     * ges_timeline_new_audio_video (void); Creates a new GESTimeline containing
     * a raw audio and a raw video track.
     *
     * Returns
     *
     * The newly created GESTimeline.
     */

    public GESTimeline(boolean audio_video) {
        this(initializer(audio_video ? ges.ptr_ges_timeline_new_audio_video() : ges.ptr_ges_timeline_new(), false));
    }

    /**
     *
     * GESTimeline * ges_timeline_new_from_uri (const gchar *uri, GError
     * **error); Creates a timeline from the given URI.
     *
     * Parameters
     *
     * uri the URI to load from error An error to be set in case something wrong
     * happens or NULL. Returns
     *
     * A new timeline if the uri was loaded successfully, or NULL if the uri
     * could not be loaded
     */

    public GESTimeline timelineFromURI(String uri) {
        Pointer[] error = { null };
        GESTimeline result = new GESTimeline(initializer(ges.ptr_ges_timeline_new_from_uri(uri, error)));
        if (null != error[0]) {
            throw new GstException(new GError(new GErrorStruct(error[0])));
        }
        return result;
    }

    /**
     * Add the layer to the timeline. The reference to the layer will be stolen
     * by the timeline .
     *
     * Parameters
     *
     * timeline a GESTimeline layer the GESLayer to add Returns
     *
     * TRUE if the layer was properly added, else FALSE.
     */
    public boolean addLayer(GESLayer layer) {
        return ges.ges_timeline_add_layer(this, layer);
    }

    /**
     * newly created GESLayer to timeline Note that you do not own any reference
     * to the returned layer.
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * The newly created GESLayer, or
     *
     * the last (empty) GESLayer of timeline .
     *
     * [transfer none]
     */
    public GESLayer appendLayer() {
        Pointer ptr = ges.ptr_ges_timeline_append_layer(this);
        GESLayer layer = new GESLayer(initializer(ptr, true, true));
        Logger.getGlobal().info(" layer refCount:" + layer.getRefCount());
        return layer;
    }

    /**
     * Removes the layer from the timeline. The reference that the timeline
     * holds on the layer will be dropped. If you wish to use the layer after
     * calling this method, you need to take a reference before calling.
     *
     * Parameters
     *
     * timeline a GESTimeline layer the GESLayer to remove Returns
     *
     * TRUE if the layer was properly removed, else FALSE.
     */

    public boolean removeLayer(GESLayer layer) {
        return ges.ges_timeline_remove_layer(this, layer);
    }

    /**
     * Add a track to the timeline. The reference to the track will be stolen by
     * the pipeline.
     *
     * Parameters
     *
     * timeline a GESTimeline track the GESTrack to add Returns
     *
     * TRUE if the track was properly added, else FALSE.
     */
    public boolean addTrack(GESTrack track) {
        return ges.ges_timeline_add_track(this, track);
    }

    /**
     * Remove the track from the timeline . The reference stolen when adding the
     * track will be removed. If you wish to use the track after calling this
     * function you must ensure that you have a reference to it.
     *
     * Parameters
     *
     * timeline a GESTimeline track the GESTrack to remove Returns
     *
     * TRUE if the track was properly removed, else FALSE.
     */

    public boolean removeTrack(GESTrack track) {
        return ges.ges_timeline_remove_track(this, track);
    }

    /**
     *
     * Loads the contents of URI into the given timeline.
     *
     * Parameters
     *
     * timeline an empty GESTimeline into which to load the formatter uri The
     * URI to load from error An error to be set in case something wrong happens
     * or NULL. Returns
     *
     * TRUE if the timeline was loaded successfully, or FALSE if the uri could
     * not be loaded.
     */
    public boolean loadFromURI(String uri) {
        Pointer[] error = { null };
        boolean result = ges.ges_timeline_load_from_uri(this, uri, error);
        if (Objects.nonNull(error[0])) {
            throw new GstException(new GError(new GErrorStruct(error[0])));
        }
        return result;
    }

    /**
     *
     *
     * Saves the timeline to the given location
     *
     * Parameters
     *
     * timeline a GESTimeline uri The location to save to formatter_asset The
     * formatter asset to use or NULL. If NULL, will try to save in the same
     * format as the one from which the timeline as been loaded or default to
     * the formatter with highest rank. overwrite TRUE to overwrite file if it
     * exists error An error to be set in case something wrong happens or NULL.
     * Returns
     *
     * TRUE if the timeline was successfully saved to the given location, else
     * FALSE.
     */

    public boolean saveToURI(String uri, GESAsset formatter_asset, boolean overwrite) {
        Pointer[] error = { null };
        boolean result = ges.ges_timeline_save_to_uri(this, uri, formatter_asset, overwrite, error);
        if (Objects.nonNull(error[0])) {
            throw new GstException(new GError(new GErrorStruct(error[0])));
        }
        return result;
    }

    /**
     *
     * Commit all the pending changes of the clips contained in the timeline .
     *
     * When changes happen in a timeline, they are not directly executed in the
     * non-linear engine. Call this method once you are done with a set of
     * changes and want it to be executed.
     *
     * The GESTimeline::commited signal will be emitted when
     *
     * the (possibly updated) GstPipeline is ready to output data again, except
     * if the state of the timeline was GST_STATE_READY or GST_STATE_NULL.
     *
     * Note that all the pending changes will automatically be executed when the
     * timeline goes from GST_STATE_READY to GST_STATE_PAUSED, which usually is
     * triggered by corresponding state changes in a containing GESPipeline.
     *
     * You should not try to change the state of the timeline, seek it or add
     * tracks to it during a commit operation, that is between a call to this
     * function and after receiving the GESTimeline::commited signal.
     *
     * See ges_timeline_commit_sync if you don't want to bother with waiting for
     * the signal.
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * TRUE if pending changes were commited or FALSE if nothing needed to be
     * commited
     */
    public boolean commit() {
        return ges.ges_timeline_commit(this);
    }

    /**
     *
     * Commit all the pending changes of the GESClips contained in the timeline
     * .
     *
     * Will return once the update is complete, that is
     *
     * when the (possibly updated) GstPipeline is ready to output data again, or
     * if the state of the timeline was GST_STATE_READY or GST_STATE_NULL.
     *
     * This function will wait for any pending state change of the timeline by
     * calling gst_element_get_state with a GST_CLOCK_TIME_NONE timeout, you
     * should not try to change the state from another thread before this
     * function has returned.
     *
     * See ges_timeline_commit for more information.
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * TRUE if pending changes were commited or FALSE if nothing needed to be
     * commited
     */
    public boolean commitSync() {
        if (GES.getVersion().getMinor() < 5 && GES.getVersion().getMajor() <=1) {
            return ges.ges_timeline_commit(this);
        } else {
            return ges.ges_timeline_commit_sync(this);
        }
    }

    /**
     *
     * Returns the list of GESTrack used by the Timeline.
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * A list of GESTrack. The caller should unref each track once he is done
     * with them.
     *
     * [transfer full][element-type GESTrack]
     */
    public List<GESTrack> getTracks() {
        List<GESTrack> list;
        GList glist = ges.ges_timeline_get_tracks(this);
        list = GList.convertToListOfType(glist, GESTrack.class, true);
        GlibAPI.GLIB_API.g_list_free(glist);
        return list;
    }

    /**
     *
     * Retrieve the layer with priority as a priority
     *
     * Parameters
     *
     * timeline The GESTimeline to retrive a layer from priority The priority of
     * the layer to find Returns
     *
     * A GESLayer or NULL if no layer with priority was found
     *
     * Since 1.6
     */
    GESLayer getLayer(int priority) {
        return ges.ges_timeline_get_layer(this, priority);
    }

    /**
     *
     * Get the list of GESLayer present in the Timeline.
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * the list of GESLayer present in the Timeline sorted by priority. The
     * caller should unref each Layer once he is done with them.
     *
     * [transfer full][element-type GESLayer]
     */
    public List<GESLayer> getLayers() {
        List<GESLayer> list;
        GList glist = ges.ges_timeline_get_layers(this);
        list = GList.convertToListOfType(glist, GESLayer.class, true);
        GlibAPI.GLIB_API.g_list_free(glist);
        return list;
    }

    /**
     *
     * Search the GESTrack corresponding to the given timeline 's pad .
     *
     * Parameters
     *
     * timeline The GESTimeline pad The GstPad Returns
     *
     * The corresponding GESTrack if it is found, or NULL if there is an error.
     *
     * [transfer none]
     */
    public GESTrack getTrackForPad(Pad pad) {
        return ges.ges_timeline_get_track_for_pad(this, pad);
    }

    /**
     *
     *
     * Search the GstPad corresponding to the given timeline 's track .
     *
     * Parameters
     *
     * - timeline The GESTimeline
     *
     * - track The GESTrack
     *
     * Returns
     *
     * The corresponding GstPad if it is found, or NULL if there is an error.
     *
     * [transfer none]
     */
    public Pad getPadForTrack(GESTrack track) {
        return ges.ges_timeline_get_pad_for_track(this, track);
    }

    /**
     *
     * Get the
     * current duration of timeline
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * The current duration of timeline
     */
    public ClockTime getDuration() {
        return ges.ges_timeline_get_duration(this);
    }

    /**
     * Gets whether transitions are automatically added when objects overlap or
     * not.
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * TRUE if transitions are automatically added, else FALSE.
     */
    public boolean getAutotransition() {
        return ges.ges_timeline_get_auto_transition(this);
    }

    /**
     *
     * Sets the layer to the given auto_transition . See the
     * documentation of the property auto_transition for more information.
     *
     * Parameters
     *
     * timeline a GESLayer auto_transition whether the auto_transition is active
     */
    public void setAutoTransition(boolean auto_transition) {
        ges.ges_timeline_set_auto_transition(this, auto_transition);
    }

    /**
     *
     *
     * Gets the configured snapping distance of the timeline. See the
     * documentation of the property snapping_distance for more information.
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * The snapping_distance property of the timeline
     */
    public ClockTime getSnappingDistance() {
        return ges.ges_timeline_get_snapping_distance(this);
    }

    /**
     *
     * Sets the snapping_distance of the timeline. See the documentation of the
     * property snapping_distance for more information.
     *
     * Parameters
     *
     * timeline a GESLayer snapping_distance whether the snapping_distance is
     * active ges_timeline_get_element ()
     */
    public void setSnappingDistance(ClockTime snaping_distance) {
        ges.ges_timeline_set_snapping_distance(this, snaping_distance);
    }

    /**
     * Gets a GESTimelineElement contained in the timeline
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * The GESTimelineElement or NULL if not found.
     *
     * [transfer full]
     */
    public GESTimelineElement getTimelineElement(String name) {
        return ges.ges_timeline_get_element(this, name);
    }
    /**
     *
     * Check whether a GESTimelineElement is empty or not
     *
     * Parameters
     *
     * timeline a GESTimeline Returns
     *
     * TRUE if the timeline is empty FALSE otherwize
     */
    public boolean isEmpty() {
        return ges.ges_timeline_is_empty(this);
    }

     /** Types and Values
     *
     * GESTimeline
     *
     * typedef struct { /* <readonly> / GList *layers; GList *tracks; }
     * GESTimeline; Members
     *
     * GList *layers; A list of GESLayer sorted by priority NOTE: Do not modify.
     * GList *tracks; A list of GESTrack sorted by priority NOTE: Do not modify.
     * Property Details
     *
     * The “auto-transition” property
     *
     * “auto-transition” gboolean Sets whether transitions are added
     * automagically when clips overlap.
     *
     * Flags: Read / Write
     *
     * Default value: FALSE
     *
     * The “duration” property
     *
     * “duration” guint64
     *
     * Current duration (in nanoseconds) of the GESTimeline
     *
     * Flags: Read
     *
     * Default value: 18446744073709551615
     *
     * The “snapping-distance” property
     *
     * “snapping-distance” guint64
     *
     * Distance (in nanoseconds) from which a moving object will snap with it
     * neighboors. 0 means no snapping.
     *
     * Flags: Read / Write
     *
     * Default value: 0
     */
}
