package org.gstreamer.ges;

import java.util.List;

import org.gstreamer.ClockTime;
import org.gstreamer.GError;
import org.gstreamer.GstException;
import org.gstreamer.discoverer.GstDiscovererInfo;
import org.gstreamer.lowlevel.GCancellable;
import org.gstreamer.lowlevel.GESUriClipAssetAPI;
import org.gstreamer.lowlevel.GType;
import org.gstreamer.lowlevel.GlibAPI.GList;
import org.gstreamer.lowlevel.GstAPI.GErrorStruct;
import org.gstreamer.lowlevel.GstAPI.GstCallback;
import org.gstreamer.lowlevel.GstNative;
import org.gstreamer.lowlevel.annotations.CallerOwnsReturn;

import com.sun.jna.Pointer;

/**
 * * GESUriClipAsset
 *
 * GESUriClipAsset — A GESAsset subclass specialized in GESUriClip extraction
 *
 * Description
 *
 * The GESUriClipAsset is a special GESAsset that lets you handle the media file
 * to use inside the GStreamer Editing Services. It has APIs that let you get
 * information about the medias. Also, the tags found in the media file are set
 * as Metadatas of the Asser.
 */

public class GESUriClipAsset extends GESClipAsset {
    private static interface API extends GESUriClipAssetAPI {
        @CallerOwnsReturn
        Pointer ptr_ges_uri_clip_asset_request_sync(String uri, Pointer[] error);
    }

    private static final API ges = GstNative.load("ges", API.class);

    public GESUriClipAsset(Initializer init) {
        super(init);
    }

    /**
     */
    public static GType assetGetType() {
        return ges.ges_uri_clip_asset_get_type();
    }

    /**
     * Gets duration of the file represented by self Parameters
     *
     * self
     *
     *
     * a GESUriClipAsset
     *
     * Returns
     *
     * The duration of self
     */

    public ClockTime getDuration() {
        return ges.ges_uri_clip_asset_get_duration(this);
    }

    /**
     * Gets Whether the file represented by self is an image or not Parameters
     *
     * self
     *
     *
     * a indent: Standard input:311: Error:Unexpected end of file
     * GESUriClipAsset
     *
     * Returns
     *
     * Whether the file represented by self is an image or not
     */
    public boolean isImage() {
        return ges.ges_uri_clip_asset_is_image(this);
    }

    /**
     * Gets GstDiscovererInfo about the file Parameters
     *
     * self
     *
     *
     * Target asset
     *
     * Returns
     *
     * GstDiscovererInfo of specified asset.
     *
     * [transfer none]
     */
    public GstDiscovererInfo getInfo() {
        return ges.ges_uri_clip_asset_get_info(this);
    }

    /**
     * Creates a GESUriClipAsset for uri
     *
     * Example of request of a GESUriClipAsset:
     *
     * // The request callback static void filesource_asset_loaded_cb (GESAsset
     * * source, GAsyncResult * res, gpointer user_data) { GError *error = NULL;
     * GESUriClipAsset *filesource_asset;
     *
     * filesource_asset = GES_URI_CLIP_ASSET (ges_asset_request_finish (res,
     * &error)); if (filesource_asset) { g_print (
     * "The file: %s is usable as a FileSource, it is%s an image and lasts %"
     * GST_TIME_FORMAT, ges_asset_get_id (GES_ASSET (filesource_asset))
     * ges_uri_clip_asset_is_image (filesource_asset) ? "" : " not",
     * GST_TIME_ARGS (ges_uri_clip_asset_get_duration (filesource_asset)); }
     * else { g_print (
     * "The file: %s is *not* usable as a FileSource because: %s",
     * ges_asset_get_id (source), error->message); }
     *
     * gst_object_unref (mfs); }
     *
     * // The request: ges_uri_clip_asset_new (uri, (GAsyncReadyCallback)
     * filesource_asset_loaded_cb, user_data);
     *
     * Parameters
     *
     * uri
     *
     *
     * The URI of the file for which to create a GESUriClipAsset
     *
     *
     * cancellable
     *
     *
     * optional GCancellable object, NULL to ignore.
     *
     *
     * callback
     *
     *
     * a GAsyncReadyCallback to call when the initialization is finished. [scope
     * async]
     *
     * user_data
     *
     *
     * The user data to pass when callback is called
     */
    public static GESUriClipAsset loadAssetAsync(String uri, GCancellable cancellable, GstCallback callback,
            Pointer user_data) {
        // TODO: implement
        return null;
    }

    /**
     * GESUriClipAsset * ges_uri_clip_asset_request_sync (const gchar *uri,
     * GError **error);
     *
     * Creates a GESUriClipAsset for uri syncronously. You should avoid to use
     * it in application, and rather create GESUriClipAsset asynchronously
     * Parameters
     *
     * uri
     *
     *
     * The URI of the file for which to create a GESUriClipAsset. You can also
     * use multi file uris for GESMultiFileSource.
     *
     *
     * error
     *
     *
     * An error to be set in case something wrong happens or NULL. [allow-none]
     * Returns
     *
     * A reference to the requested asset or NULL if an error happend.
     *
     *
     * [transfer none]
     */
    public static GESUriClipAsset requestSync(String uri) {
        Pointer[] error = { null };
        ges.ges_uri_clip_asset_request_sync(uri, error);
        Pointer ptr = ges.ptr_ges_uri_clip_asset_request_sync(uri, error);
        if (null != error[0]) {
            throw new GstException(new GError(new GErrorStruct(error[0])));
        }
        GESUriClipAsset result = new GESUriClipAsset(initializer(ptr, true, true));
        // GESUriClipAsset result = ges.ges_uri_clip_asset_request_sync(uri,
        // error);
        return result;
    }

    /**
     * Get the GESUriSourceAsset self containes Parameters
     *
     * self
     *
     *
     * A GESUriClipAsset
     *
     * Returns
     *
     * a GList of GESUriSourceAsset.
     *
     * [transfer none][element-type GESUriSourceAsset]
     */
    public List<GESUriSourceAsset> getStreamAssets() {
        GList glist = ges.ges_uri_clip_asset_get_stream_assets(this);
        List<GESUriSourceAsset> list = GList.convertToListOfType(glist, GESUriSourceAsset.class);
        return list;
    }

    /**
     *
     * Sets the timeout of GESUriClipAsset loading Parameters
     *
     * klass
     *
     *
     * The GESUriClipAssetClass on which to set the discoverer timeout
     *
     *
     * timeout
     *
     *
     * The timeout to set
     *
     * Types and Values GESUriClipAsset
     *
     * typedef struct _GESUriClipAsset GESUriClipAsset;
     */
    public static void classSetTtimeout(GESUriClipAssetClass klass, ClockTime timeout) {
        // TODO: implement
    }
}
