package org.gstreamer.ges.clip;

import java.util.List;
import java.util.Objects;

import org.gstreamer.ges.GESAsset;
import org.gstreamer.ges.GESContainer;
import org.gstreamer.ges.GESLayer;
import org.gstreamer.ges.GESTrack;
import org.gstreamer.ges.GESTrackElement;
import org.gstreamer.ges.GESTrackType;
import org.gstreamer.ges.operation.GESBaseEffect;
import org.gstreamer.lowlevel.GESClipAPI;
import org.gstreamer.lowlevel.GType;
import org.gstreamer.lowlevel.GlibAPI.GList;
import org.gstreamer.lowlevel.GstTypes;

/**
 * GESClip
 *
 * GESClip — Base Class for objects in a GESLayer
 *
 * Description
 *
 * A GESClip is a 'natural' object which controls one or more GESTrackElement(s)
 * in one or more GESTrack(s).
 *
 * Keeps a reference to the GESTrackElement(s) it created and sets/updates their
 * properties.
 *
 */
public class GESClip extends GESContainer {
    public static final String GTYPE_NAME = "GESClip";

    private static GESClipAPI ges = GESClipAPI.GSTCLIP_API;

    public GESClip(Initializer init) {
        super(init);
    }

    /**
     * Get the GESLayer to which this clip belongs. Parameters
     *
     * clip
     *
     *
     * a GESClip
     *
     * Returns
     *
     * The GESLayer where this clip is being used, or NULL if it is not used on
     * any layer. The caller should unref it usage.
     *
     * [transfer full]
     */
    public GESLayer getLayer() {
        return ges.ges_clip_get_layer(this);
    }

    /**
     *
     * Finds the GESTrackElement controlled by clip that is used in track . You
     * may optionally specify a GType to further narrow search criteria.
     *
     * Note: If many objects match, then the one with the highest priority will
     * be returned. Parameters
     *
     * clip
     *
     *
     * a GESClip
     *
     *
     * track
     *
     *
     * a GESTrack or NULL. [allow-none]
     *
     * type
     *
     *
     * a GType indicating the type of track element you are looking for or
     * G_TYPE_NONE if you do not care about the track type.
     *
     * Returns
     *
     * The GESTrackElement used by track , else NULL, Unref after usage.
     *
     * [transfer full]
     */
    public <T extends GESTrackElement> T findTrackElement(GESTrack track, Class<T> gesTrackElementType) {
        GType type = GType.NONE;
        if (Objects.nonNull(gesTrackElementType)) {
            type = GstTypes.typeFor(gesTrackElementType);
        }

        return (T) ges.ges_clip_find_track_element(this, track, type);
    }

    public <T extends GESTrackElement> T findTrackElement(GESTrack track) {
        return (T) ges.ges_clip_find_track_element(this, track, GType.NONE);
    }
    
     /**

     *
     * Finds all the GESTrackElement controlled by clip that is used in track .
     * You may optionally specify a GType to further narrow search criteria.
     * Parameters
     *
     * clip
     *
     *
     * a GESClip
     *
     *
     * track
     *
     *
     * a GESTrack or NULL. [allow-none]
     *
     * track_type
     *
     *
     * a GESTrackType indicating the type of tracks in which elements should be
     * searched.
     *
     *
     * type
     *
     *
     * a GType indicating the type of track element you are looking for or
     * G_TYPE_NONE if you do not care about the track type.
     *
     * Returns
     *
     * a GList of the GESTrackElement contained in clip . The refcount of the
     * objects will be increased. The user will have to unref each
     * GESTrackElement and free the GList.
     *
     * [transfer full][element-type GESTrackElement]
     */
    public List<GESTrackElement> findTrackElements(GESTrack track, GESTrackType track_type) {
        return findTrackElements(track,track_type,GESTrackElement.class);
    }

    public <T extends GESTrackElement> List<T> findTrackElements(GESTrack track, GESTrackType track_type, Class<T> gesTrackElementType) {
        GType type = GType.NONE;
        Objects.requireNonNull(gesTrackElementType);
        if (Objects.nonNull(gesTrackElementType)) {
            type = GstTypes.typeFor(gesTrackElementType);
        }

        GList glist = ges.ges_clip_find_track_elements(this, track, track_type, type);

        return GList.convertToListOfType(glist, gesTrackElementType, true);
    }

     /**
     * Extracts a GESTrackElement from asset and adds it to the clip . Should
     * only be called in order to add operations to a GESClip, ni other cases
     * TrackElement are added automatically when adding the GESClip/GESAsset to
     * a layer.
     *
     * Takes a reference on track_element . Parameters
     *
     * clip
     *
     *
     * a GESClip
     *
     *
     * asset
     *
     *
     * a GESAsset with GES_TYPE_TRACK_ELEMENT as extractable_type
     *
     * Returns
     *
     * Created GESTrackElement or NULL if an error happened.
     *
     * [transfer none][allow-none]
     */
    public GESTrackElement addAsset (GESAsset asset) {
        return ges.ges_clip_add_asset(this, asset);
    }

     /**
     * Get effects applied on clip Parameters
     *
     * clip
     *
     *
     * The origin GESClip
     *
     * Returns
     *
     * a GList of the GESBaseEffect that are applied on clip order by ascendant
     * priorities. The refcount of the objects will be increased. The user will
     * have to unref each GESBaseEffect and free the GList.
     *
     * [transfer full][element-type GESTrackElement]
     */
    public List<GESTrackElement> getTopEffects() {
        GList glist = ges.ges_clip_get_top_effects (this);
        return GList.convertToListOfType(glist, GESTrackElement.class);
    }

     /**
     *
     * Gets the index position of an effect. Parameters
     *
     * clip
     *
     *
     * The origin GESClip
     *
     *
     * effect
     *
     *
     * The GESBaseEffect we want to get the top index from
     *
     * Returns
     *
     * The top index of the effect, -1 if something went wrong.
     */
    public int getTopEffectIndex (GESBaseEffect effect) {
        return ges.ges_clip_get_top_effect_index(this, effect);
    }


     /**
     * Moves clip to layer . If clip is not in any layer, it adds it to layer ,
     * else, it removes it from its current layer, and adds it to layer .
     * Parameters
     *
     * clip
     *
     *
     * a GESClip
     *
     *
     * layer
     *
     *
     * the new GESLayer
     *
     * Returns
     *
     * TRUE if clip could be moved FALSE otherwize
     */
    public boolean moveToLayer (GESLayer layer) {
        return ges.ges_clip_move_to_layer(this, layer);
    }


     /**
      *  This is a convenience method that lets you set the index of a top effect.
     * Parameters
     *
     * clip
     *
     *
     * The origin GESClip
     *
     *
     * effect
     *
     *
     * The GESBaseEffect to move
     *
     *
     * newindex
     *
     *
     * the new index at which to move the effect inside this GESClip
     *
     * Returns
     *
     * TRUE if effect was successfuly moved, FALSE otherwise.
     */
     public boolean setTopEffectIndex (GESBaseEffect effect, int newindex) {
         return ges.ges_clip_set_top_effect_index(this, effect, newindex);
     }

     /**
     * Sets the formats supported by the file. Parameters
     *
     * clip
     *
     *
     * the GESClip to set supported formats on
     *
     *
     * supportedformats
     *
     *
     * the GESTrackType defining formats supported by clip
     */
     public void setSupportedFormats (GESTrackType supportedformats) {
         ges.ges_clip_set_supported_formats(this, supportedformats);
     }


     /**
     * Get the formats supported by clip . Parameters
     *
     * clip
     *
     *
     * the GESClip
     *
     * Returns
     *
     * The formats supported by clip .
     */
     public int getSupportedFormats () {
         return ges.ges_clip_get_supported_formats(this);
     }

     /**
     *
     * The function modifies clip , and creates another GESClip so we have two
     * clips at the end, splitted at the time specified by position , as a
     * position in the timeline (not in the clip to be split). For example, if
     * ges_clip_split is called on a 4-second clip playing from 0:01.00 until
     * 0:05.00, with a split position of 0:02.00, this will result in one clip
     * of 1 second and one clip of 3 seconds, not in two clips of 2 seconds.
     *
     * The newly created clip will be added to the same layer as clip is in.
     * This implies that clip must be in a GESLayer for the operation to be
     * possible.
     *
     * This method supports clips playing at a different tempo than one second
     * per second. For example, splitting a clip with a GESEffect 'pitch
     * tempo=1.5' four seconds after it starts, will set the inpoint of the new
     * clip to six seconds after that of the clip to split. For this, the
     * rate-changing property must be registered using
     * ges_effect_class_register_rate_property ; for the 'pitch' plugin, this is
     * already done. Parameters
     *
     * clip
     *
     *
     * the GESClip to split
     *
     *
     * position
     *
     *
     * a GstClockTime representing the timeline position at which to split
     *
     * Returns
     *
     * The newly created GESClip resulting from the splitting.
     *
     * [transfer none]
     */
     public GESClip clipSplit (long position) {
         return ges.ges_clip_split (this, position);
     }

     /*
     * Types and Values GESClip
     *
     * typedef struct _GESClip GESClip;
     *
     * The GESClip base class. GESClipClass
     *
     * typedef struct { GESCreateTrackElementFunc create_track_element;
     * GESCreateTrackElementsFunc create_track_elements; } GESClipClass;
     *
     * Subclasses can override the create_track_element . Members
     *
     * GESCreateTrackElementFunc create_track_element;
     *
     *
     * method to create a single GESTrackElement for a given GESTrack.
     *
     *
     * GESCreateTrackElementsFunc create_track_elements;
     *
     *
     * method to create multiple GESTrackElements for a GESTrack.
     *
     * Property Details The “layer” property
     *
     * “layer” GESLayer *
     *
     * The GESLayer where this clip is being used. If you want to connect to its
     * notify signal you should connect to it with g_signal_connect_after as the
     * signal emission can be stop in the first fase.
     *
     * Flags: Read The “supported-formats” property
     *
     * “supported-formats” GESTrackType
     *
     * The formats supported by the clip.
     *
     * Flags: Read / Write / Construct
     *
     * Default value: GES_TRACK_TYPE_AUDIO | GES_TRACK_TYPE_VIDEO
     */
}
